
/********************************************************
** calendar.c
**
** Part of Calendar Project, Copyright (c) 2004-2005 [Yak]
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2.1 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
********************************************************/

/* Week number calculations modified by NuukiaWorld in 2006
   stamp2d() date calculation fixed by NuukiaWorld in 2007
*/

#include "calendar.h"

#define NUM_OF_ROWS        (5)
#define DAYS_PER_SCREEN    (NUM_OF_ROWS*7)

#ifndef  HIRES
// Default display (84 x 48) layout
#define  BOXW  12
#define  BOXH  6
#define  AREAX 0
#define  AREAY 1
#else
// Hires display layout
#define  BOXW  13
#define  BOXH  9
#define  AREAX 3
#define  AREAY 1
#endif   // HIRES

/* structure created in RAM upon calendar opening */
struct calendar
{
   struct calendar *prev;                                            // points to previous if more than one calendar is opened
   
   unsigned (*handler)(unsigned mode, unsigned *udata, void *arg);   // external handler; tells if day should be bolded
   unsigned udata;                                                   // external handlers userdata field

   unsigned start, curr;                                             // start of screen view, currently selected day
   unsigned drawn_start, drawn_curr;                                 // same as above but for last redraw
   unsigned drawn_today;
   long drawn_weekstart;                                             // weekstart value used for last redraw
   unsigned char drawn_weekend;
};

/* external data */
extern nokstr_t str_week;
extern const UBYTE own_font_data[];

/* REXX vars transfer */
extern rexx_uint confid_week_start;
extern rexx_uint confid_weekend;

extern rexx_uint confid_first_week;

/* tracks various data when calendar is opened */
struct calendar *calendar;

/* days per month, used by date manipulation routines */
static const unsigned char dayspermonth[]={31,28,31,30,31,30,31,31,30,31,30,31};

/************************************************************************************************/

void draw_day(char *s, unsigned x, unsigned y, unsigned dmode);

/************************************************************************************************/
/* own date manipulation routines */

static unsigned stamp2d(long stamp) /* converts system timestamp to our date */
{
   /*     2050/1/1     secs/day */
//   return 748396+((stamp+1)/86400);
   /* System timestamp is a signed number of seconds since 2050/1/1 0:00:00.
      First add offset (21550 * 86400) to the timestamp so that it will be an unsigned number of
      seconds since 1991/1/1.
      Then divive by number of seconds per day (86400) to get number of days since 1991/1/1.
      Then add the value representing 1991/1/1 in our date system (726847).
   */
   return (unsigned)( stamp + 21550 * 86400 ) / 86400 + 726847;
}

static unsigned date2d(const struct date *date) /* converts system date struct to our date */
{
   unsigned d=0, m;

   if(date->year!=0) d=(((date->year-1)/4)*1461)+(((date->year-1)%4)*365);
   if(date->day!=0) d+=date->day-1;
 
   if(date->month!=0)
   {
      for(m=0;m<date->month-1;m++)
      {
         d+=dayspermonth[m];
         if((m==1)&&(date->year%4==0)) d++; // february
      }
   }

   return d;
}

static void d2date(struct date *date, unsigned d) /* converts our date to system date struct */
{
   unsigned r, m;

   date->year=(d/1461)*4;
   r=(d%1461)/365;
   if(r==4)
   {
      date->year+=4;
      date->month=12;
      date->day=31;
      return;
   }
   date->year+=r+1;
   d=(d%1461)%365;

   for(m=0;m<12;m++)
   {
      r=dayspermonth[m];
      if((m==1)&&(date->year%4==0)) r++;
      if(d<r) break;
      d-=r;
   }
   date->month=m+1;
   date->day=d+1;
}

static unsigned dweekday(unsigned d) /* returns weekday of our date (0=sunday) */
{
   return d%7;
}

/* returns year week of our date
   firstweek = 0 (A: starts on Jan 1), 3 (B: first 4-day week), or 6 (C: first full week)
*/
static ULONG dweek( unsigned int d, int weekstart, unsigned int firstweek )
{
   unsigned base;

   // find the A:7th, B:4th, or C:1st day of the week
   // it tells the year whose week numbering should be used
   d = ( d - weekstart ) / 7 * 7 + weekstart + 6 - firstweek;
   // find the A:1st, B:4th, or C:7th Jan of that year, it defines the week #1
   base = d % 1461;                             // 1461 = 365 + 365 + 365 + 366
   base = ( base < 1460 ? base % 365 : 366 );   // = day of year
   base = d - base + firstweek;
   // Return the difference of these two weeks plus one
   return ( d - weekstart ) / 7 - ( base - weekstart ) / 7 + 1;
}

static unsigned dlimit(unsigned d) /* limits our date to a range of 1991/1/1 - 2090/12/31 */
{
   if(d<726847) d=726847; /* 1991/1/1 */
   else if(d>763371) d=763371; /* 2090/12/31 */
   return d;
}

/************************************************************************************************/

/* opens calendar screen (does not draw it; calendar_set_date() should be used to do it) */
void calendar_open(unsigned (*handler)(unsigned mode, unsigned *udata, void *arg))
{
   struct calendar *c;
   
   c=malloc(sizeof(struct calendar));
   memset(c,0,sizeof(struct calendar));
   c->handler=handler;
// c->start=0x7fffffff; // so when opening, selected date is visible at the top op screen

   freegfx_open_ex(FGF_ENABLE_BLINK|FGF_SHOW_ACTION);

   c->prev=calendar;
   calendar=c;
}

/************************************************************************************************/

/* closes calendar screen and frees it's resources */
void calendar_close(void)
{
   struct calendar *c;
   
   c=calendar;
   if(!c) return;

   calendar=c->prev;
   free(c);
   freegfx_close();
}

/************************************************************************************************/

#ifndef HIRES
void draw_day(char *s, unsigned x, unsigned y, unsigned dmode)
{
   struct bitmap b;

   memset(&b,0,sizeof(struct bitmap));
   b.width=4; b.height=5;
   y-=4; // base
   while(*s)
   {
      if((*s>='0')&&(*s<='9'))
      {
         b.data=(UBYTE *)(own_font_data+(*s-'0')*4);
         own_draw_bitmap(&b,x,y,(dmode==DM_BOLD)?DM_SET:dmode);
         if(dmode==DM_BOLD) own_draw_bitmap(&b,x-1,y,DM_OR);
      }
      s++;
      x+=5;
   }
}
#else
void draw_day(char *s, unsigned x, unsigned y, unsigned dmode)
{
   const struct font *font;

   if(dmode==DM_BOLD)
   {
      font=own_get_font("tiny/bold");
      dmode=DM_SET;
      x--;
   }
   else font=own_get_font("tiny/plain");

   if(s[0]==' ')
   {
      x+=own_get_string_width((const nokstr_t *)"0",font);
      s++;
   }

   own_print_string((const nokstr_t *)s,x,y,font,dmode);
}
#endif   // HIRES

/************************************************************************************************/

/* draws/redraws the calendar screen (set full to TRUE if handler MUST be called for all visible days) */
void calendar_draw(bool full)
{
   struct date date; /* first so its ptr%4==0 */
   unsigned d,day,today;
   const struct font *pfont, *bfont;
   int x, y, i, weekstart, weekend, firstweek, dm;
   unsigned int areay;  // int is more effective on ARM7
   char sbuf[10];
   wchar_t wbuf[30];
   wchar_t *wptr;

   if(!calendar) return;

   today=stamp2d(get_timestamp());
   weekstart=own_get_config(confid_week_start)-1;
   if(weekstart>1) weekstart=1; /* use default if invalid weekstart */

   firstweek = own_get_config(confid_first_week) * 3;
   if(weekstart>6) weekstart=3; /* use default if invalid firstweek */

   weekend=own_get_config(confid_weekend);

   /* this modifies the calendar->start when weekstart is changed */
   if(calendar->start>0) calendar->start+=weekstart-calendar->drawn_weekstart;

   /* make sure calendar->curr will be visible */
   x=dweekday(calendar->curr)-weekstart;
   if(x<0) x+=7;
   if(x>=7) x-=7;
   d=calendar->curr-x;
 
   /* When the calendar opens first time, show the current day on the 2nd row */
   if( calendar->start == 0 ) calendar->start = d - 7;
 
   else if(calendar->curr<calendar->start) calendar->start=d;
   else if(calendar->curr>=calendar->start+DAYS_PER_SCREEN) calendar->start=d-DAYS_PER_SCREEN+7;

   /* size of the day box */
//   if(screen_info.width==84) { BOXW=12; BOXH=6; AREAX=0; areay=1; }
//   else { BOXW=13; BOXH=9; AREAX=3; areay=1; }

   pfont=own_get_font("small/plain");
   bfont=own_get_font("small/bold");
   areay=AREAY+own_get_char_height('A',bfont);

   /* decide if we should perform full screen redraw or only cursor movement without scroll */
   if((full)||(calendar->start!=calendar->drawn_start)||(calendar->drawn_weekstart!=weekstart)||(calendar->drawn_weekend!=weekend))
   {
      full=TRUE;
      freegfx_begin(TRUE);

      /* draw weekend separators */
      i=weekstart+1-weekend;
      x=AREAX+((6*BOXW-i*BOXW-1)%(7*BOXW));
      if((x>=AREAX)&&(x<=AREAX+6*BOXW)) own_draw_line(x,areay,x,areay+NUM_OF_ROWS*BOXH-2,1,DM_SET);
      x=AREAX+((8*BOXW-i*BOXW-1)%(7*BOXW));
      if((x>=AREAX)&&(x<=AREAX+6*BOXW)) own_draw_line(x,areay,x,areay+NUM_OF_ROWS*BOXH-2,1,DM_SET);
   }
   else
   {
      freegfx_begin(FALSE);
      own_draw_box(0,0,screen_info.width,own_get_char_height('A',bfont),DM_CLEAR); /* clear status line */
   }

   if((full)&&(calendar->handler)) calendar->handler(CHM_INIT,&calendar->udata,NULL);

   x=AREAX;
   y=areay;
   for(d=0;d<DAYS_PER_SCREEN;d++)
   {
      day=calendar->start+d;

      if(full)
      {
         d2date(&date,day);

         dm=DM_SET;
         if(calendar->handler)
         {
            if(calendar->handler(CHM_CHECK_DATE,&calendar->udata,&date)) dm=DM_BOLD;
         }
         
         ntoa2(date.day,sbuf);
         if(sbuf[0]=='0') sbuf[0]=' ';
         draw_day(sbuf,x+1,y+BOXH-2,dm);
      }
      else
      {
         if(day==calendar->drawn_curr) own_draw_box(x,y,BOXW-1,BOXH,DM_INVERT);
         if(day==calendar->drawn_today) own_draw_box(x,y,BOXW-1,BOXH,DM_NONE);
      }

      if(day==calendar->curr)
      {
         own_draw_box(x,y,BOXW-1,BOXH,DM_INVERT);

         /* month and year */
         d2date(&date,day);
         wcscpy(wbuf,w_get_string(&str_short_months[date.month-1],current_language));
         i=wcslen(wbuf);
         wbuf[i++]=' ';
         ntoa(date.year,sbuf);
         wcscpy(wbuf+i,w_get_string((const nokstr_t *)sbuf,0));
         wcsmark(wbuf);
         own_print_string((const nokstr_t *)wbuf,
            screen_info.width-own_get_string_width((const nokstr_t *)wbuf,pfont),
            own_get_char_base('A',pfont),pfont,DM_SET);

         /* weekday and year week */
         wptr=(WCSPTR)&str_short_weekdays[dweekday(calendar->curr)];
         own_print_string((const nokstr_t *)wptr,0,
            own_get_char_base('A',bfont),bfont,DM_SET);
   
         own_print_string((const nokstr_t *)w_sprintf((const nokstr_t *)"/%N",
            dweek( calendar->curr, weekstart, firstweek ) ),
            own_get_string_width((const nokstr_t *)wptr,bfont),
            own_get_char_base('A',pfont),pfont,DM_SET);
      }
  
      if(day==today) own_draw_box(x,y,BOXW-1,BOXH,DM_BLINK);

      if(x==AREAX+6*BOXW) { x=AREAX; y+=BOXH; } /* end of line, go to start of next */
      else x+=BOXW; /* next item in line */
   }

   if((full)&&(calendar->handler)) calendar->handler(CHM_EXIT,&calendar->udata,NULL);

   freegfx_end(FALSE); /* end refreshing (no immediate screen update) */

   calendar->drawn_start=calendar->start;
   calendar->drawn_curr=calendar->curr;
   calendar->drawn_today=today;
   calendar->drawn_weekstart=weekstart;
   calendar->drawn_weekend=weekend;
}

/************************************************************************************************/

/* handles the keypresses (call it when keypress arives) */
bool calendar_keypress(unsigned key)
{
   struct date date;
   unsigned curr,i,j;
   long diff;

   if(!calendar) return false;

   curr=calendar->curr;
   d2date(&date,curr);

   switch(key)
   {
      case KEY_1: curr-=8; break;
      case KEY_2: curr-=7; break;
      case KEY_3: curr-=6; break;
      case KEY_4: curr-=1; break;
      case KEY_6: curr+=1; break;
      case KEY_7: curr+=6; break;
      case KEY_8: curr+=7; break;
      case KEY_9: curr+=8; break;

      case KEY_HASH:
         curr=stamp2d(get_timestamp());
         calendar->curr = calendar->start = 0; // Show the current day on the second row
         break;

      case KEY_ASTERISK: // year--
         if((date.year%4==0)&&(date.month>2)) diff=366;
         else if(((date.year-1)%4==0)&&(date.month<=2)) diff=366;
         else diff=365;
         curr-=diff;
         break;

      case KEY_0: // year++
         if((date.year%4==0)&&(date.month<=2)) diff=366;
         else if(((date.year+1)%4==0)&&(date.month>2)) diff=366;
         else diff=365;
         curr+=diff;
         break;

      case KEY_UP: // month--
         if(--date.month<1) { date.year--; date.month=12; }
         j=dayspermonth[date.month-1];
         if((date.month==2)&&(date.year%4==0)) j++;
         if(date.day>j) j=date.day;
         curr-=j;
         break;

      case KEY_DOWN: // month++
         i=dayspermonth[date.month-1];
         if((date.month==2)&&(date.year%4==0)) i++;
         if(++date.month>12) { date.year++; date.month=1; }
         j=dayspermonth[date.month-1];
         if((date.month==2)&&(date.year%4==0)) j++;
         if(date.day>j) i+=j-date.day;
         curr+=i;
         break;
   }
 
   if(curr!=calendar->curr)
   {
      curr=dlimit(curr);
      if(curr!=calendar->curr)
      {
         calendar->curr=curr;
         calendar_draw(false);
         return true;
      }
   }

   return false;
}

/************************************************************************************************/

/* used to set current date selected in calendar (also redraws the screen)*/
void calendar_set_date(const struct date *date, bool full)
{
   if(calendar)
   {
      calendar->curr=dlimit(date2d(date));
      calendar_draw(full);
   }
}

/************************************************************************************************/

/* used to get current date selected in calendar */
void calendar_get_date(struct date *date)
{
   if(calendar) d2date(date,calendar->curr);
}

/************************************************************************************************/
